package de.killig.qbist;

import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import android.app.ProgressDialog;
import android.app.WallpaperManager;
import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Rect;
import android.media.MediaScannerConnection;
import android.os.Environment;
import android.os.Handler;
import android.util.AttributeSet;
import android.util.Log;
import android.view.MotionEvent;
import android.view.View;
import android.widget.Toast;

import static android.view.MotionEvent.*;

class QbistView extends View {
	private Bitmap[] mBitmap=new Bitmap[9];
//  private boolean[] mBitmapDirty=new boolean[9];
  private long mStartTime;
  private static final String LOG_TAG = "QbistView";
  private static final int NUM_TRANSFORMS=36;
  int W, H;
	final static int position[][] = {{1,1},{0,0},{1,0},{2,0},{2,1},{2,2},{1,2},{0,2},{0,1}};
  //Thread t;
	MyThreadGroup mtg=new MyThreadGroup("QbistViewThreadGroup");
  ProgressDialog progressDialog;
  String mMessage;
//  float pressureMax;
	enum EMode {VARIATIONS, WALLPAPER, SAVE};
//	EMode mode;
  static final int AVAILABLE_PROCESSORS=/*2*/Runtime.getRuntime().availableProcessors(); // Differs when called later?!
	
  private static native void renderQbist(Bitmap bitmap, long time_ms, int gVariation);
  private static native void variationAsMain(int gVariation, short coarseness);
  private static native void MakeVariations(short coarseness);
  static native void OnNewDocument();
//  static native void SwitchCoarseness(short newCoarseness);
  private static native void Serialize(byte[] ar);

  final Handler mHandler = new Handler();
  final Runnable mUpdateResults = new Runnable() {
    public void run() {
			if(progressDialog!=null) progressDialog.dismiss();
			if(mMessage!=null) {
				Toast.makeText(getContext(), mMessage, Toast.LENGTH_LONG).show();
				mMessage=null;
			}
    }
  };


/*  public QbistView(Context context, int width, int height) {
      super(context);
      W = width;
      H = height;
      mStartTime = System.currentTimeMillis();
    	OnNewDocument();
      init();
  }*/
  
  public QbistView(Context context, AttributeSet attrs) {
		super(context, attrs);
		Log.d(LOG_TAG, "QbistView(Context context, AttributeSet attrs)");
	}

  public QbistView(Context context) {
		this(context, null);
	}

  
  void init() {
  	//OnNewDocument();
		for(int gVariation=0; gVariation<9; gVariation++) {
      mBitmap[gVariation] = Bitmap.createBitmap(W/3, H/3, Bitmap.Config.ARGB_8888);
      /*renderQbist(mBitmap[gVariation], System.currentTimeMillis() - mStartTime, gVariation);
			mBitmapDirty[gVariation]=true;
      invalidate();*/
		}
		renderQbistThread(mBitmap, System.currentTimeMillis() - mStartTime, 0, 8, EMode.VARIATIONS);
  }
  

  @Override
  protected void onLayout(boolean changed, int left, int top, int right, int bottom) {
  	super.onLayout(changed, left, top, right, bottom);
		Log.d(LOG_TAG, "onLayout("+changed);
    W = right-left;
    H = bottom-top;
/*    if(changed) {
    	OnNewDocument();
      init();
    }*/
  }
  
  class QbistViewThread extends Thread {
  	public Bitmap[] bitmap;
  	long time_ms;
  	EMode mode=null;
		public int gVariationFirst;
		public int gVariationLast;

    public QbistViewThread(ThreadGroup tg, String string) {
			super(tg, string);
		}

		public void run() {
  		for(int gVariation=gVariationFirst; gVariation<=gVariationLast && !isInterrupted(); gVariation++) {
  			renderQbist(bitmap[gVariation], time_ms, gVariation);
  			switch(mode) {
  			case VARIATIONS:
//			 				mBitmapDirty[gVariation]=true;
	 				postInvalidate(position[gVariation][0]*W/3,position[gVariation][1]*H/3, position[gVariation][0]*W/3+W/3, position[gVariation][1]*H/3+H/3);
 					break;
  			case WALLPAPER:
 		    	try {
 		      	WallpaperManager.getInstance(getContext()).setBitmap(bitmap[0]);
		    			mMessage=null;
 					} catch (IOException e) {
 						e.printStackTrace();
 						mMessage=e.getMessage();
 					}
 					break;
	  		case SAVE:
		    		if (Environment.MEDIA_MOUNTED.equals(Environment.getExternalStorageState())) {
	 		    	try {
 		    			File path = Environment.getExternalStoragePublicDirectory(Environment.DIRECTORY_PICTURES);
 		    			path.mkdirs();
 		  				File f;
 		  				int i=0;
 		  				do {
 		  					f=new File(path, "Qbist"+(++i)+".png");
 		  				} while(f.exists());
 		  				
 		          bitmap[0].compress(Bitmap.CompressFormat.PNG, 0, new FileOutputStream(f));
 		          MediaScannerConnection.scanFile(getContext(), new String[]{f.getAbsolutePath()}, null, null);
 		          byte[] ar=new byte[NUM_TRANSFORMS*4];
 		          Serialize(ar);
 		          File f2=new File(path, "Qbist"+i+".qbe");
 		          DataOutputStream dos=new DataOutputStream(new FileOutputStream(f2));
 		          for(byte s : ar) {
 		          	dos.writeByte(0);
 		          	dos.writeByte(s);
 		          }
 		          dos.close();
 		          mMessage=f.getAbsolutePath()+" + .qbe";
	 					} catch (IOException e) {
	 						e.printStackTrace();
	 						mMessage=e.getMessage();
	 					}
		    		} else {
 						mMessage="SD Card not writable!";
		    		}
 					break;
  			}
					mHandler.post(mUpdateResults);
  		}

    	
    }
  }
  
  
  void renderQbistThread(final Bitmap[] bitmap,final long time_ms, final int gVariationFirst, final int gVariationLast, final EMode mode) {
    /*Thread t = new Thread(tg, "QbistViewThread") {
      public void run() {*/
      	int noVariations=gVariationLast-gVariationFirst+1;
    		Log.d(LOG_TAG, "AVAILABLE_PROCESSORS="+AVAILABLE_PROCESSORS);
      	int noThreads=Math.min(AVAILABLE_PROCESSORS, noVariations);
//noThreads=2;
    		Log.d(LOG_TAG, "noThreads="+noThreads);
    		QbistViewThread[] t2=new QbistViewThread[noThreads];//(gVariationLast-gVariationFirst)/
      	final int variationsPerThread=noVariations/noThreads;
    		Log.d(LOG_TAG, "variationsPerThread="+variationsPerThread);
      	int gVariationFirstThread;
      	//int gVariationLastThread;
      	gVariationFirstThread=gVariationFirst;
      	for(int thread=0; thread<noThreads /*&& !isInterrupted()*/; thread++) {
      		t2[thread]=new QbistViewThread(mtg,"QbistViewThread"+thread);
      		t2[thread].bitmap=bitmap;
      		t2[thread].time_ms=time_ms;
      		t2[thread].mode=mode;
      		t2[thread].gVariationLast=Math.min(gVariationFirstThread+variationsPerThread, gVariationLast);
      		t2[thread].gVariationFirst=gVariationFirstThread;
      		t2[thread].start();
        	gVariationFirstThread+=variationsPerThread;
      	}
      	/*for(int thread=0; thread<noThreads; thread++) {
      		try {
						t2[thread].join();
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
      	}
      }
    };
    t.start();*/
  }
  

  @Override
  public boolean onTouchEvent(MotionEvent event) {
//  	pressureMax=Math.max(pressureMax, event.getPressure());
  	switch(event.getAction()) {
  	case ACTION_DOWN:
//  		pressureMax=0;
//	    mStartTime = event.getEventTime();
			Log.d(LOG_TAG, "ACTION_DOWN  x:"+event.getX()+" y:"+event.getY()/*+" pressureMax="+pressureMax+" mStartTime="+mStartTime*/);
			return true; // else we don't get ACTION_UP, in contrast to a dedicated OnTouchListener (?!)
  	case ACTION_UP:
			long time=event.getEventTime() - mStartTime;
    	if(mtg!=null /*&& tg.isAlive()*/) {
    		mtg.interrupt();
    		//try {
					mtg.join();
				/*} catch (InterruptedException e) {
					e.printStackTrace();
				}*/
    	}
    	float xe=event.getX();
    	float ye=event.getY();
			Log.d(LOG_TAG, "ACTION_UP  x:"+xe+" y:"+ye/*+" pressureMax="+pressureMax+" time="+time*/);
			int pos;
//xe=799.7417f;
			int xr=(int)(xe/((float)W/3f));
			int yr=(int)(ye/((float)H/3f));
			for(pos=0; pos<position.length; pos++) {
				if(xr==position[pos][0] && yr==position[pos][1]) {
					float l=position[pos][0]*W/3;
					//float t=position[pos][1]*H/3;
					float lc=(xe-l)/(W/3);
					//float tc=(ye-l/*really l?*/)/(H/3);
					float c=lc*8;
					Log.d(LOG_TAG, "pos="+pos+" c="+c);
					short coarseness=(short)/*Math.max(1.0f, ( (pressureMax)*3.0f*(time/100.0f) ) )*/c;
					if(pos==0) MakeVariations(coarseness);
					else variationAsMain(pos, coarseness);
					break;
				}
			}
			mBitmap[0]=Bitmap.createBitmap(mBitmap[pos]);
//			mBitmapDirty[0]=true;
  		/*for(int gVariation=1; gVariation<9; gVariation++) {
  			renderQbist(mBitmap[gVariation], System.currentTimeMillis() - mStartTime, gVariation);
 				mBitmapDirty[gVariation]=true;
        invalidate();
  		}*/
  		invalidate(position[0][0]*W/3,position[0][1]*H/3, position[0][0]*W/3+W/3, position[0][1]*H/3+H/3);
  		renderQbistThread(mBitmap, System.currentTimeMillis() - mStartTime, 1, 8, EMode.VARIATIONS);
			return true;
  	default:
  		return super.onTouchEvent(event);
  	}
  }
  
  
  /*void onResume() {
 		for(int gVariation=0; gVariation<9; gVariation++) {
			if(mBitmap[gVariation]!=null) mBitmapDirty[gVariation]=true;
      invalidate(position[gVariation][0]*W/3,position[gVariation][1]*H/3, position[gVariation][0]*W/3+W/3, position[gVariation][1]*H/3+H/3);
		}
  }*/

  @Override
  protected void onDraw(Canvas canvas) {
  	Rect r=canvas.getClipBounds();
		Log.d(LOG_TAG, "onDraw() "+r);
		if(mBitmap[0]==null) {
    	OnNewDocument();
      init();
		}
 		for(int gVariation=0; gVariation<9; gVariation++) {
 			//if(mBitmapDirty[gVariation]) {
 			if(Rect.intersects(r, new Rect(position[gVariation][0]*W/3,position[gVariation][1]*H/3, position[gVariation][0]*W/3+W/3, position[gVariation][1]*H/3+H/3))) {
 				Log.d(LOG_TAG, "drawBitmap("+gVariation);
 				canvas.drawBitmap(mBitmap[gVariation], position[gVariation][0]*W/3, position[gVariation][1]*H/3, null);
 				//mBitmapDirty[gVariation]=false;
 			}
		}
		super.onDraw(canvas);
  }
  

  void save(boolean wallpaper) {
    Bitmap[] bitmap=new Bitmap[1];
  	WallpaperManager m=WallpaperManager.getInstance(getContext());
  	int w=m.getDesiredMinimumWidth();
  	int h=m.getDesiredMinimumHeight();
  	if(w<=0 || h<=0) {
  		w=W;
  		h=H;
  	}
  	bitmap[0] = Bitmap.createBitmap(w, h, Bitmap.Config.ARGB_8888);
  	//renderQbist(bitmap, System.currentTimeMillis() - mStartTime, 0);
  	progressDialog = ProgressDialog.show(getContext(), null, "Rendering...", true);
//  	mode=wallpaper?EMode.WALLPAPER:EMode.SAVE;
  	renderQbistThread(bitmap, System.currentTimeMillis() - mStartTime, 0, 0, wallpaper?EMode.WALLPAPER:EMode.SAVE);
  	/*try {
			t.join();
		} catch (InterruptedException e1) {
			e1.printStackTrace();
		}
  	
  	try {
			File f;
			int i=0;
			do {
				f=new File(Environment.getExternalStorageDirectory().toString(), "Qbist"+(++i)+".png");
			} while(f.exists());
      bitmap[0].compress(Bitmap.CompressFormat.PNG, 0, new FileOutputStream(f));
      byte[] ar=new byte[NUM_TRANSFORMS*4];
      Serialize(ar);
      File f2=new File(Environment.getExternalStorageDirectory().toString(), "Qbist"+i+".qbe");
      DataOutputStream dos=new DataOutputStream(new FileOutputStream(f2));
      for(byte s : ar) {
      	dos.writeByte(0);
      	dos.writeByte(s);
      }
      dos.close();
			m.setStream(new FileInputStream(f));
      Toast.makeText(getContext(), f.getAbsolutePath()+"/.qbe", Toast.LENGTH_LONG).show();
      
      progressDialog.dismiss();
      
		} catch (IOException e) {
			e.printStackTrace();
			Toast.makeText(getContext(), e.getMessage(), Toast.LENGTH_LONG).show();
		}*/
  }
}
